unit im_Effects;

{*********************************************}
{  This unit is a part of ImageE              }
{  Copyright  2003-2004 R.Geurts             }
{  See Readme.txt for licence information     }
{Effect_Emboss
Effect_Gradient
}
{*********************************************}

interface

uses
  sysutils,graphics,GR32, GR32_Filters,math,GR32_Blend,
  im_D5,
  Windows, Messages, Classes, Controls, Forms, Dialogs;

  // Color Conversion
  function HSLtoRGB(H, S, L: Integer): TColor32;
  procedure RGBtoHSL(RGB: TColor32; out H, S, L: Byte);

procedure Effect_Dither(src: TBitmap32);

function ArcTan2(xt,yt : Single): Single;
procedure Effect_addmononoise(clip: tbitmap32; Amount: Integer);
procedure Effect_addcolornoise(clip: tbitmap32; Amount: Integer);
procedure Effect_Saturation(clip: tbitmap32; Amount: Integer);
procedure Effect_FilterRed(clip: tbitmap32; min,max: Integer);
procedure Effect_XFilterRed(clip: tbitmap32; min,max: Integer);
procedure Effect_FilterBlue(clip: tbitmap32; min,max: Integer);
procedure Effect_XFilterBlue(clip: tbitmap32; min,max: Integer);
procedure Effect_FilterGreen(clip: tbitmap32; min,max: Integer);
procedure Effect_XFilterGreen(clip: tbitmap32; min,max: Integer);
procedure Effect_Posterize(src: tbitmap32; amount: integer);
procedure Effect_Solarize(src: tbitmap32; amount: integer);
procedure Effect_Emboss(src: tbitmap32);
procedure Effect_Splitblur(clip: tbitmap32; Amount: Integer);
procedure Effect_GaussianBlur(clip: tbitmap32; Amount: integer);
procedure Effect_ToBinary(src: tbitmap32; amount: integer);
procedure Effect_Mosaic(Bm:TBitmap32;size:Integer);
procedure Effect_IncDecRGB(clip:tbitmap32;r0,g0,b0:integer);
procedure Effect_IncDecHSL(clip:tbitmap32;h0,s0,l0:integer);
procedure Effect_WaveH(clip:tbitmap32;amount,inference,style:integer);
procedure Effect_WaveV(clip:tbitmap32;amount,inference,style:integer);

procedure Effect_Twist(Bmp: TBitmap32; Amount: integer);
procedure Effect_FishEye(bmp: TBitmap32; Amount: Extended);
procedure Effect_Addborder(bmp:tbitmap32; new_width,new_height: integer);
procedure Effect_ScrollXY(bmp: tbitmap32; AmountX,AmountY: Integer);

procedure Effect_ConvertToPolar(bmp: tbitmap32);
procedure Effect_Gradient(bmp: tbitmap32; min,max,operation:integer);


const
  MaxPixelCount = 32768;

implementation

type
    pRGBTripleArray = ^TRGBTripleArray;
    TRGBTripleArray = ARRAY[0..MaxPixelCount-1] OF TRGBTriple;

 pRGBArray = ^TRGBArray; // Use SysUtils.pByteArray
 TRGBArray = ARRAY[0..MaxPixelCount-1] OF TRGBTriple;

function IntToByte(i:Integer):Byte;
begin
  if      i>255 then Result:=255
  else if i<0   then Result:=0
  else               Result:=i;
end;

procedure Effect_Addborder(bmp:tbitmap32; new_width,new_height: integer);
var
temp:tbitmap32;
temp2:tbitmap32;
xoff,yoff:integer;
begin
if new_width<bmp.Width then exit;
if new_height<bmp.height then exit;
  temp := TBitMap32.create;
  temp2:= tbitmap32.create;
  temp2.Assign(bmp);

  temp.Width:=new_width;
  temp.height:=new_height;
  temp.fillrect(0,0,temp.width,temp.height,clblack);
  bmp.Assign(temp);

 xoff:=round((bmp.Width-temp2.width)/2);
 yoff:=round((bmp.height-temp2.height)/2);
 bmp.Draw(xoff,yoff,temp2);
 temp.Free;
 temp2.free;
end;


procedure Effect_Emboss(src: tbitmap32);
var x,y:integer;
    p1,p2:Pcolor32array;
    r1,g1,b1,r0,g0,b0, r,g,b:Word;//:integer;
    SS:Pcolor32;
begin
  SS := @src.Bits[0];
  for y:=0 to src.height-2 do begin
    p1:=src.scanline[y];
    p2:=src.scanline[y+1];
   for x:=0 to src.width-1 do begin
      b0:=(p1[x] and $FF);
      g0:=(p1[x] shr 8) and $FF;
      r0:=(p1[x] shr 16) and $FF;
      b1:=(p2[x]and $FF);
      g1:=(p2[x] shr 8) and $FF;;
      r1:=(p2[x] shr 16) and $FF;;
       b:=(r0+(r1 xor $FFFFFFFF)and $FF)shr 1;
       g:=(g0+(g1 xor $FFFFFFFF)and $FF)shr 1;
       r:=(b0+(b1 xor $FFFFFFFF)and $FF)shr 1;
      if r>255 then r:=255;
      if g>255 then g:=255;
      if b>255 then b:=255;
      SS^ := $FF000000 + r shl 16 + g shl 8 + b;
    inc(SS);
    end;
   end;
   src.Changed;
end;

procedure Effect_WaveH(clip:tbitmap32;amount,inference,style:integer);
var
  x,y : integer;
  BitMap : TBitMap32;
  P1,P2 : Pcolor32array;
  fangle:real;
  b:integer;
begin
if amount=0 then exit;
//if (inference<1) or (inference>50) then exit;
if (style<0) or (style>2) then exit;
  BitMap := TBitMap32.create;
  Bitmap.assign(clip);
  fangle:=pi / 2 / amount;
    for y := BitMap.height -1-(2*amount) downto amount do begin
      P1 := BitMap.ScanLine[y];
      b:=0;
      for x:=0 to Bitmap.width-1 do begin
        P2 := clip.scanline[y+amount+b];
       p2[x]:=p1[x];
        case style of
        0: b:=amount*variant(sin(fangle*x));
        1: b:=amount*variant(sin(fangle*x)*cos(fangle*x));
        2: b:=amount*variant(sin(fangle*x)*sin(inference*fangle*x));
        end;
      end;
    end;
  BitMap.free;
  clip.changed;
end;

procedure Effect_WaveV(clip:tbitmap32;amount,inference,style:integer);
var
  x,y : integer;
  BitMap : TBitMap32;
  P1,P2 : Pcolor32array;
  fangle:real;
  x2:integer;
begin
  BitMap := TBitMap32.create;
  Bitmap.assign(clip);
  fangle:=pi / 2 / amount;
    for y := BitMap.height-1 downto 0 do begin
    for x:=(2*amount) to Bitmap.width-1-amount do begin
      P1 := BitMap.ScanLine[y];
      P2 := clip.scanline[y];

      x2 := x+round( amount*variant(sin(fangle*inference)));
      p2[x]:=p1[x2];
      end;
      end;

  BitMap.free;
  clip.changed;
end;

  function ArcTan2(xt,yt : Single): Single;
  begin
    if xt = 0 then
      if yt > 0 then
        Result := Pi/2
      else
        Result := -(Pi/2)
    else begin
      Result := ArcTan(yt/xt);
      if xt < 0 then
        Result := Pi + ArcTan(yt/xt);
    end;
  end;

procedure Effect_Twist(Bmp: TBitmap32; Amount: integer);
var
  fxmid, fymid : Single;
  txmid, tymid : Single;
  fx,fy : Single;
  tx2, ty2 : Single;
  rt : Single;
  theta : Single;
  ifx, ify : integer;
  dx, dy : Single;
  OFFSET : Single;
  ty, tx             : Integer;
  weight_x, weight_y     : array[0..1] of Single;
  weight                 : Single;
  new_red, new_green     : Integer;
  new_blue               : Integer;
  total_red, total_green : Single;
  total_blue             : Single;
  ix, iy                 : Integer;
  sli, slo : Pcolor32Array;
  dst : TBitMap32;
  SS:Pcolor32;
  r,g,b,d:word;//integer;

begin
if amount=0 then exit;
   SS := @bmp.Bits[0];

  dst := TBitMap32.create;
  dst.assign(bmp);

  OFFSET := -(Pi/2);
  dx := Bmp.Width - 1;
  dy := Bmp.Height - 1;
  rt := Sqrt(dx * dx + dy * dy);
  tx2 := rt;
  ty2 := rt;
  txmid := (Bmp.Width-1)/2;    //Adjust these to move center of rotation
  tymid := (Bmp.Height-1)/2;   //Adjust these to move ......
  fxmid := (Bmp.Width-1)/2;
  fymid := (Bmp.Height-1)/2;
  if tx2 >= Bmp.Width then tx2 := Bmp.Width-1;
  if ty2 >= Bmp.Height then ty2 := Bmp.Height-1;

  for ty := 0 to Round(ty2) do begin
    for tx := 0 to Round(tx2) do begin
      dx := tx - txmid;
      dy := ty - tymid;
      rt := Sqrt(dx * dx + dy * dy);
      if rt = 0 then begin
        fx := 0;
        fy := 0;
      end
      else begin
        theta := ArcTan2(dx,dy) - rt/Amount - OFFSET;
        fx := rt * Cos(theta);
        fy := rt * Sin(theta);
      end;
      fx := fx + fxmid;
      fy := fy + fymid;

      ify := Trunc(fy);
      ifx := Trunc(fx);
                // Calculate the weights.
      if fy >= 0  then begin
        weight_y[1] := fy - ify;
        weight_y[0] := 1 - weight_y[1];
      end else begin
        weight_y[0] := -(fy - ify);
        weight_y[1] := 1 - weight_y[0];
      end;
      if fx >= 0 then begin
        weight_x[1] := fx - ifx;
        weight_x[0] := 1 - weight_x[1];
      end else begin
        weight_x[0] := -(fx - ifx);
        Weight_x[1] := 1 - weight_x[0];
      end;

      if ifx < 0 then
        ifx := Bmp.Width-1-(-ifx mod Bmp.Width)
      else if ifx > Bmp.Width-1  then
        ifx := ifx mod Bmp.Width;
      if ify < 0 then
        ify := Bmp.Height-1-(-ify mod Bmp.Height)
      else if ify > Bmp.Height-1 then
        ify := ify mod Bmp.Height;

      total_red   := 0.0;
      total_green := 0.0;
      total_blue  := 0.0;
      for ix := 0 to 1 do begin
        for iy := 0 to 1 do begin
          if ify + iy < Bmp.Height then
            sli := Bmp.scanline[ify + iy]
          else
            sli := Bmp.scanline[Bmp.Height - ify - iy];
          if ifx + ix < Bmp.Width then begin
      new_red:=round(sli[ifx+ix] and $FF);
      new_green:=round((sli[ifx+ix] shr 8) and $FF);
      new_blue:=round((sli[ifx+ix] shr 16) and $FF);
{            new_red := sli[(ifx + ix)*3];
            new_green := sli[(ifx + ix)*3+1];
            new_blue := sli[(ifx + ix)*3+2];
}          end
          else begin
      new_red:=round(sli[bmp.width-ifx-ix] and $FF);
      new_green:=round((sli[bmp.width-ifx-ix] shr 8) and $FF);
      new_blue:=round((sli[bmp.width-ifx-ix] shr 16) and $FF);
{            new_red := sli[(Bmp.Width - ifx - ix)*3];
            new_green := sli[(Bmp.Width - ifx - ix)*3+1];
            new_blue := sli[(Bmp.Width - ifx - ix)*3+2];
}          end;
          weight := weight_x[ix] * weight_y[iy];
          total_red   := total_red   + new_red   * weight;
          total_green := total_green + new_green * weight;
          total_blue  := total_blue  + new_blue  * weight;
        end;
      end;
      slo := Dst.scanline[ty];
      r := Round(total_red);
      g := Round(total_green);
      b := Round(total_blue);
      slo[tx]:= $FF000000 + b shl 16 + g shl 8 + r;
    end;
  end;
  for ty:=0 to dst.Height-1 do
  begin
    slo:=dst.scanline[ty];
    for tx:=0 to dst.Width-1 do
    begin
      b:=(slo[tx] and $FF);
      g:=(slo[tx] shr 8) and $FF;
      r:=(slo[tx] shr 16) and $FF;
    SS^ := $FF000000 + r shl 16 + g shl 8 + b;
    inc(SS);
    end;
   end;
 bmp.Changed;
end;



procedure Effect_FishEye(bmp: TBitmap32; Amount: Extended);
var
xmid,ymid              : Single;
fx,fy                  : Single;
r1, r2                 : Single;
ifx, ify               : integer;
dx, dy                 : Single;
rmax                   : Single;
ty, tx                 : Integer;
weight_x, weight_y     : array[0..1] of Single;
weight                 : Single;
new_red, new_green     : Integer;
new_blue               : Integer;
total_red, total_green : Single;
total_blue             : Single;
ix, iy                 : Integer;
sli, slo : Pcolor32Array;
  dst : TBitMap32;
  SS:Pcolor32;
  r,g,b,d:word;//integer;
  tmp:extended;
begin
  amount:=amount/10;
  if amount=0 then exit;
   SS := @bmp.Bits[0];

  dst := TBitMap32.create;
  dst.assign(bmp);

  xmid := bmp.Width/2;
  ymid := bmp.Height/2;
  rmax := Dst.Width * Amount;
  for ty := 0 to Dst.Height - 1 do begin
    for tx := 0 to Dst.Width - 1 do begin
      dx := tx - xmid;
      dy := ty - ymid;
      r1 := Sqrt(dx * dx + dy * dy);
      if r1 = 0 then begin
        fx := xmid;
        fy := ymid;
      end
      else begin

        if rmax=0 then rmax:=1;
        tmp:=(1-r1/rmax)-1;
        if tmp=0 then tmp:=1;

        r2 := rmax/2 * (1 / tmp);
        fx := (dx * r2 / r1 + xmid);
        fy := (dy * r2 / r1 + ymid);

      end;
      ify := Trunc(fy);
      ifx := Trunc(fx);
                // Calculate the weights.
      if fy >= 0  then begin
        weight_y[1] := fy - ify;
        weight_y[0] := 1 - weight_y[1];
      end else begin
        weight_y[0] := -(fy - ify);
        weight_y[1] := 1 - weight_y[0];
      end;
      if fx >= 0 then begin
        weight_x[1] := fx - ifx;
        weight_x[0] := 1 - weight_x[1];
      end else begin
        weight_x[0] := -(fx - ifx);
        Weight_x[1] := 1 - weight_x[0];
      end;

      if ifx < 0 then
        ifx := bmp.Width-1-(-ifx mod bmp.Width)
      else if ifx > bmp.Width-1  then
        ifx := ifx mod bmp.Width;
      if ify < 0 then
        ify := bmp.Height-1-(-ify mod bmp.Height)
      else if ify > bmp.Height-1 then
        ify := ify mod bmp.Height;

      total_red   := 0.0;
      total_green := 0.0;
      total_blue  := 0.0;
      for ix := 0 to 1 do begin
        for iy := 0 to 1 do begin
          if ify + iy < bmp.Height then
            sli := bmp.scanline[ify + iy]
          else
            sli := bmp.scanline[bmp.Height - ify - iy];
          if ifx + ix < bmp.Width then begin

      new_blue:=round(sli[ifx+ix] and $FF);
      new_green:=round((sli[ifx+ix] shr 8) and $FF);
      new_red:=round((sli[ifx+ix] shr 16) and $FF);
          end
          else begin
      new_blue:=round(sli[bmp.width-ifx-ix] and $FF);
      new_green:=round((sli[bmp.width-ifx-ix] shr 8) and $FF);
      new_red:=round((sli[bmp.width-ifx-ix] shr 16) and $FF);
          end;
          weight := weight_x[ix] * weight_y[iy];
          total_red   := total_red   + new_red   * weight;
          total_green := total_green + new_green * weight;
          total_blue  := total_blue  + new_blue  * weight;
        end;
      end;
      slo := dst.scanline[ty];
      r := Round(total_red);
      g := Round(total_green);
      b := Round(total_blue);
      slo[tx]:= $FF000000 + r shl 16 + g shl 8 + b;
//    SS^ := $FF000000 + r shl 16 + g shl 8 + b;
    end;
  end;
//
  for ty:=0 to dst.Height-1 do
  begin
    slo:=dst.scanline[ty];
    for tx:=0 to dst.Width-1 do
    begin
  {    b:=(slo[tx] and $FF);
      g:=(slo[tx] shr 8) and $FF;
      r:=(slo[tx] shr 16) and $FF;
    SS^ := $FF000000 + r shl 16 + g shl 8 + b;
   } SS^:=slo[tx];
    inc(SS);
    end;
   end;
   dst.Free;
   bmp.Changed;
end;

procedure Effect_GaussianBlur(clip: tbitmap32; Amount: integer);
var
i: Integer;
begin
  for i:=Amount downto 0 do
  Effect_SplitBlur(clip,1);
end;

procedure Effect_Splitblur(clip: tbitmap32; Amount: Integer);
var
//p0,
p1,p2:Pcolor32array;
r,g,b:word;
cx,x,y: Integer;
Buf:   array[0..3,0..2]of integer;
SS:Pcolor32;
begin
  SS := @clip.Bits[0];

  if Amount=0 then Exit;
  for y:=0 to clip.Height-1 do
  begin
//    p0:=clip.scanline[y];
    if y-Amount<0         then p1:=clip.scanline[y]
    else {y-Amount>0}          p1:=clip.ScanLine[y-Amount];
    if y+Amount<clip.Height    then p2:=clip.ScanLine[y+Amount]
    else {y+Amount>=Height}    p2:=clip.ScanLine[clip.Height-y];

    for x:=0 to clip.Width-1 do
    begin
      if x-Amount<0     then cx:=x
      else {x-Amount>0}      cx:=x-Amount;

      Buf[0,0]:=(p1[cx] and $FF);
      Buf[0,1]:=(p1[cx] shr 8) and $FF;
      Buf[0,2]:=(p1[cx] shr 16) and $FF;
      Buf[1,0]:=(p2[cx] and $FF);
      Buf[1,1]:=(p2[cx] shr 8) and $FF;
      Buf[1,2]:=(p2[cx] shr 16) and $FF;
      if x+Amount<clip.Width     then cx:=x+Amount
      else {x+Amount>=Width}     cx:=clip.Width-x;
      Buf[2,0]:=(p1[cx] and $FF);
      Buf[2,1]:=(p1[cx] shr 8) and $FF;
      Buf[2,2]:=(p1[cx] shr 16) and $FF;
      Buf[3,0]:=(p2[cx] and $FF);
      Buf[3,1]:=(p2[cx] shr 8) and $FF;
      Buf[3,2]:=(p2[cx] shr 16) and $FF;

      r:=(Buf[0,0]+Buf[1,0]+Buf[2,0]+Buf[3,0])shr 2;
      g:=(Buf[0,1]+Buf[1,1]+Buf[2,1]+Buf[3,1])shr 2;
      b:=(Buf[0,2]+Buf[1,2]+Buf[2,2]+Buf[3,2])shr 2;

      SS^ := $FF000000 + b shl 16 + g shl 8 + r;
      inc(SS);

    end;
  end;
     clip.Changed;
end;




procedure Effect_addmononoise(clip: tbitmap32; Amount: Integer);
var
p0:Pcolor32array;
x,y: Integer;
r,g,b,a: word;
SS:Pcolor32;
begin
  SS := @clip.Bits[0];

  for y:=0 to clip.Height-1 do
  begin
    p0:=clip.scanline[y];
    for x:=0 to clip.Width-1 do
    begin
      a:=Random(Amount)-(Amount shr 1);
      b:=inttobyte((p0[x] and $FF)+a);
      g:=inttobyte((p0[x] shr 8) and $FF+a);
      r:=inttobyte((p0[x] shr 16) and $FF+a);

    SS^ := $FF000000 + r shl 16 + g shl 8 + b;
    inc(SS);
    end;
  end;
  clip.Changed;
end;

procedure Effect_addcolornoise(clip: tbitmap32; Amount: Integer);
var
p0:Pcolor32array;
Toggle:Boolean;
Amount1,
x,y: Integer;
//a,
AmountW, r,g,b: word;
///h1,h2,h3,h4:integer;
SS:Pcolor32;
begin
  SS := @clip.Bits[0];

  for y:=0 to clip.Height-1 do
  begin
    p0:=clip.scanline[y];
    for x:=0 to clip.Width-1 do
    begin
    {Need to change to getting variable. Check > 0 and use as a word}
      Amount1:=(Random(Amount)-(Amount shr 1));
      Toggle:= (Amount1 < 0);
      If Toggle then
      begin
         AmountW:=abs(Amount1);
      b:=inttobyte((p0[x] and $FF)- AmountW);
      end
      else
      begin
        AmountW:=abs(Amount1);
      b:=inttobyte((p0[x] and $FF)+ AmountW);
      end;

      Amount1:=(Random(Amount)-(Amount shr 1));
      Toggle:= (Amount1 < 0);
      If Toggle then
      begin
         AmountW:=abs(Amount1);
      g:=inttobyte((p0[x]  shr 8) and $FF - AmountW);
      end
      else
      begin
        AmountW:=abs(Amount1);
      g:=inttobyte((p0[x] shr 8) and $FF + AmountW);
      end;

      Amount1:=(Random(Amount)-(Amount shr 1));
      Toggle:= (Amount1 < 0);
      If Toggle then
      begin
         AmountW:=abs(Amount1);
      r:=inttobyte((p0[x]  shr 16) and $FF- AmountW);
      end
      else
      begin
        AmountW:=abs(Amount1);
      r:=inttobyte((p0[x]  shr 16) and $FF + AmountW);
      end;

{      b:=inttobyte((p0[x] and $FF)+ (Random(Amount)-(Amount shr 1)));
      g:=inttobyte((p0[x] shr 8) and $FF+(Random(Amount)-(Amount shr 1)));
      r:=inttobyte((p0[x] shr 16) and $FF+(Random(Amount)-(Amount shr 1)));}

    SS^ := $FF000000 + r shl 16 + g shl 8 + b;
    inc(SS);
    end;
  end;
  clip.Changed;
end;

procedure Effect_Posterize(src: tbitmap32; amount: integer);
var {w,h,}x,y:integer;
{,pd}    ps:Pcolor32array;
//    c:integer;
    r,g,b:word;//integer;
    SS:Pcolor32;
//    dst:tbitmap32;
begin
if amount=0 then exit;
  SS := @src.Bits[0];

  for y:=0 to src.height-1 do begin
   ps:=src.scanline[y];
   for x:=0 to src.width-1 do begin
      b:=round((((ps[x] and $FF))/amount)*amount);
      g:=round(((ps[x] shr 8) and $FF)/amount)*amount;
      r:=round(((ps[x] shr 16) and $FF)/amount)*amount;
      SS^ := $FF000000 + r shl 16 + g shl 8 + b;
    inc(SS);
    end;
   end;
   src.Changed;
end;

procedure Effect_Solarize(src: tbitmap32; amount: integer);
var x,y:integer;
    ps{,pd}:Pcolor32array;
    c:integer;
    r0,g0,b0:integer;
    r,g,b:word;
    SS:Pcolor32;
    //dst:tbitmap32;
begin
  SS := @src.Bits[0];

  for y:=0 to src.height-1 do begin
   ps:=src.scanline[y];
   for x:=0 to src.width-1 do begin
      b0:=(ps[x] and $FF);
      g0:=(ps[x] shr 8) and $FF;
      r0:=(ps[x] shr 16) and $FF;
      c:=(r0+g0+b0)div 3;
    if c>amount then begin
      b:=255-b0;
      g:=255-g0;
      r:=255-r0;
      end
      else begin
      b:=b0;
      g:=g0;
      r:=r0;
      end;
      SS^ := $FF000000 + r shl 16 + g shl 8 + b;
    inc(SS);
    end;
   end;
   src.Changed;
end;

procedure Effect_ToBinary(src: tbitmap32; amount: integer);
var
//w,h,
x,y:integer;
    ps{,pd}:Pcolor32array;
    c:integer;
    r0,g0,b0:integer;
    r:word;
    SS:Pcolor32;
    //dst:tbitmap32;
begin
  SS := @src.Bits[0];

  for y:=0 to src.height-1 do begin
   ps:=src.scanline[y];
   for x:=0 to src.width-1 do begin
      b0:=(ps[x] and $FF);
      g0:=(ps[x] shr 8) and $FF;
      r0:=(ps[x] shr 16) and $FF;
      c:=(r0+g0+b0)div 3;
      if (c>amount) then r:=$ff
    else r:=$00;

      SS^ := $FF000000 + r shl 16 + r shl 8 + r;
    inc(SS);
    end;
   end;
   src.Changed;
end;

procedure Effect_Mosaic(Bm:TBitmap32;size:Integer);
var
   x,y,i,j:integer;
   p1{,p2}:Pcolor32array;
   r,g,b:byte;
   SS:Pcolor32;
//   h1,h2,h3:integer;
begin
  SS := @bm.Bits[0];

  y:=0;
  repeat
    p1:=bm.scanline[y];
//    x:=0;
    repeat
      j:=1;
      repeat
//      p2:=bm.scanline[y];
      x:=0;
      repeat
        b:=(p1[x] and $FF);
        g:=(p1[x] shr 8) and $FF;
        r:=(p1[x] shr 16) and $FF;
        i:=1;
       repeat
    SS^ := $FF000000 + r shl 16 + g shl 8 + b;
    inc(SS);
       inc(x);
       inc(i);
       until (x>=bm.width) or (i>size);
      until x>=bm.width;
      inc(j);
      inc(y);
      until (y>=bm.height) or (j>size);
    until (y>=bm.height) or (x>=bm.width);
  until y>=bm.height;
end;



procedure Effect_FilterRed(clip: tbitmap32; min,max: Integer);
var
p0:Pcolor32array;
//g,b,a
x,y: Integer;
r: word;
SS:Pcolor32;
begin
  SS := @clip.Bits[0];
  for y:=0 to clip.Height-1 do
  begin
    p0:=clip.scanline[y];
    for x:=0 to clip.Width-1 do
    begin
//      b:=inttobyte((p0[x] and $FF));
//      g:=inttobyte((p0[x] shr 8) and $FF);
      r:=inttobyte((p0[x] shr 16) and $FF);
//      if (b>min) and (b<max) then b:=$ff else b:=0;
      SS^ := $FF000000 + r shl 16 + 0 shl 8 + 0;
    inc(SS);
    end;
  end;
  clip.Changed;
end;

procedure Effect_XFilterRed(clip: tbitmap32; min,max: Integer);
var
p0:Pcolor32array;
x,y{,a,r,g,b}: Integer;
r,g,b: word;
SS:Pcolor32;
begin
  SS := @clip.Bits[0];
  for y:=0 to clip.Height-1 do
  begin
    p0:=clip.scanline[y];
    for x:=0 to clip.Width-1 do
    begin
      b:=inttobyte((p0[x] and $FF));
      g:=inttobyte((p0[x] shr 8) and $FF);
      r:=inttobyte((p0[x] shr 16) and $FF);
      if (b>min) and (b<max) then b:=$ff
      else b:=0;
      SS^ := $FF000000 + b shl 16 + g shl 8 + r;
    inc(SS);
    end;
  end;
  clip.Changed;
end;

procedure Effect_FilterGreen(clip: tbitmap32; min,max: Integer);
var
p0:Pcolor32array;
x,y{,g,a,r,b}: Integer;
g:Word;
SS:Pcolor32;
begin
  SS := @clip.Bits[0];
  for y:=0 to clip.Height-1 do
  begin
    p0:=clip.scanline[y];
    for x:=0 to clip.Width-1 do
    begin
//      r:=inttobyte((p0[x] and $FF));
      g:=inttobyte((p0[x] shr 8) and $FF);
//      b:=inttobyte((p0[x] shr 16) and $FF);
      if (g>min) and (g<max) then g:=$ff
      else g:=0;
      SS^ := $FF000000 + 0 shl 16 + g shl 8 + 0;
    inc(SS);
    end;
  end;
  clip.Changed;
end;

procedure Effect_XFilterGreen(clip: tbitmap32; min,max: Integer);
var
p0:Pcolor32array;
x,y{,a,r,g,b}: Integer;
r,g,b: word;
SS:Pcolor32;
begin
  SS := @clip.Bits[0];
  for y:=0 to clip.Height-1 do
  begin
    p0:=clip.scanline[y];
    for x:=0 to clip.Width-1 do
    begin
      r:=inttobyte((p0[x] and $FF));
      g:=inttobyte((p0[x] shr 8) and $FF);
      b:=inttobyte((p0[x] shr 16) and $FF);
      if (g>min) and (g<max) then g:=$ff
      else g:=0;
      SS^ := $FF000000 + b shl 16 + g shl 8 + r;
    inc(SS);
    end;
  end;
  clip.Changed;
end;

procedure Effect_FilterBlue(clip: tbitmap32; min,max: Integer);
var
p0:Pcolor32array;
x,y{a,,g,br}: Integer;
r:word;
SS:Pcolor32;
begin
  SS := @clip.Bits[0];
  for y:=0 to clip.Height-1 do
  begin
    p0:=clip.scanline[y];
    for x:=0 to clip.Width-1 do
    begin
      r:=inttobyte((p0[x] and $FF));
//      g:=inttobyte((p0[x] shr 8) and $FF);
 //     b:=inttobyte((p0[x] shr 16) and $FF);
      if (r>min) and (r<max) then r:=$ff
      else r:=0;

      SS^ := $FF000000 + 0 shl 16 + 0 shl 8 + r;
    inc(SS);
    end;
  end;
  clip.Changed;
end;

procedure Effect_XFilterBlue(clip: tbitmap32; min,max: Integer);
var
p0:Pcolor32array;
x,y{,a,r,g,b}: Integer;
r,g,b:word;
SS:Pcolor32;
begin
  SS := @clip.Bits[0];
  for y:=0 to clip.Height-1 do
  begin
    p0:=clip.scanline[y];
    for x:=0 to clip.Width-1 do
    begin
      r:=inttobyte((p0[x] and $FF));
      g:=inttobyte((p0[x] shr 8) and $FF);
      b:=inttobyte((p0[x] shr 16) and $FF);
      if (r>min) and (r<max) then r:=$ff
      else r:=0;
      SS^ := $FF000000 + b shl 16 + g shl 8 + r;
    inc(SS);
    end;
  end;
  clip.Changed;
end;


procedure Effect_Saturation(clip: tbitmap32; Amount: Integer);
var
p0:Pcolor32array;
x,y{,a,r,g,b}: Integer;
r,g,b:word;
gray{,h4}:integer;
SS:Pcolor32;
begin
  SS := @clip.Bits[0];

  for y:=0 to clip.Height-1 do
  begin
    p0:=clip.scanline[y];
    for x:=0 to clip.Width-1 do
    begin
      b:=(p0[x] and $FF);
      g:=(p0[x] shr 8) and $FF;
      r:=(p0[x] shr 16) and $FF;
      Gray:=(r+g+b)div 3;
      r:=IntToByte(Gray+(((r-Gray)*Amount)div 255));
      g:=IntToByte(Gray+(((g-Gray)*Amount)div 255));
      b:=IntToByte(Gray+(((b-Gray)*Amount)div 255));

    SS^ := $FF000000 + r shl 16 + g shl 8 + b;
    inc(SS);
    end;
  end;
  clip.Changed;
end;

procedure Effect_ScrollXY(bmp: tbitmap32; AmountX,AmountY: Integer);
var
p0:Pcolor32array;
x,y{,a,r,g,b}: Integer;
r,g,b:word;
SS:Pcolor32;
x1,y1:integer;
temp:tbitmap32;
begin
//  SS := @bmp.Bits[0];

  temp:=tbitmap32.create;
  temp.assign(bmp);

  for y:=0 to temp.Height-1 do
  begin
    p0:=temp.scanline[y];
    for x:=0 to temp.Width-1 do
    begin
      b:=(p0[x] and $FF);
      g:=(p0[x] shr 8) and $FF;
      r:=(p0[x] shr 16) and $FF;

    x1:=x+amountx;
    y1:=y+amounty;

    if x1>temp.width-1 then x1:=x1-temp.Width;
    if y1>temp.height-1 then y1:=y1-temp.height;

    x1:=ensurerangeiiii(x1,0,temp.width-1);
    y1:=ensurerangeiiii(y1,0,temp.height-1);
    SS:=bmp.PixelPtr[x1,y1];
    SS^ := $FF000000 + r shl 16 + g shl 8 + b;
//    inc(SS);
    end;
  end;
  temp.free;
  bmp.Changed;
end;

procedure Effect_Gradient(bmp: tbitmap32; min,max,operation:integer);
var
//p0:Pcolor32array;
x,y{,r,a,g,b}: Integer;
r:word;
//gray,h4:integer;
SS:Pcolor32;
//x1,y1:integer;
temp:tbitmap32;
sg{,tmpi}:integer;
//mnh,mnw,theta,rt,  ,tmpr
ti,tj,fx,fy:extended;
begin
  SS := @bmp.Bits[0];

  temp:=tbitmap32.create;
  temp.assign(bmp);

  sg:=1;
  if max<min then
  begin
  sg:=-1;
{  tmpi:=max;
  max:=min;
  min:=tmpi;
}  end;

  for y:=0 to temp.Height-1 do
  begin
//    p0:=temp.scanline[y];
    for x:=0 to temp.Width-1 do
    begin
{      b:=(p0[x] and $FF);
      g:=(p0[x] shr 8) and $FF;
      r:=(p0[x] shr 16) and $FF;
}

//range 0..1
ti:=(x/(temp.width-1));
tj:=(y/(temp.height-1));
if sg=-1 then
begin
  ti:=(((temp.width-1)-x)/(temp.width-1));
  tj:=(((temp.height-1)-y)/(temp.height-1));
end;
   fy:=0;   fx:=0; {shut up compiler}
case operation of
0:  begin
      fx:=ti;
      fy:=0;
    end;
1:  begin
      fx:=0;
      fy:=tj;
    end;
2:  begin
      fx:=ti/2;
      fy:=tj/2;
    end;
3:  begin
    if ti<0.5 then fx:=ti/2;
    if ti>=0.5 then fx:=(1-ti)/2;
    if tj<0.5 then fy:=tj/2;
    if tj>=0.5 then fy:=(1-tj)/2;
    end;
end;

    r:=round((fx+fy)*(max-min)+min);

    SS^ := $FF000000 + r shl 16 + r shl 8 + r;
    inc(SS);
    end;
  end;
  temp.free;
  bmp.Changed;
end;




procedure Effect_ConvertToPolar(bmp: tbitmap32);
var
p0:Pcolor32array;
x,y{a,r,g,b}: Integer;
r,g,b:word;
//gray,h4:integer;
SS:Pcolor32;
x1,y1:integer;
temp:tbitmap32;
dx,dy:single;
//zmod,rho,
rt,theta:extended;
begin
//  SS := @bmp.Bits[0];

  temp:=tbitmap32.create;
  temp.assign(bmp);

  for y:=0 to temp.Height-1 do
  begin
    p0:=temp.scanline[y];
    for x:=0 to temp.Width-1 do
    begin
      b:=(p0[x] and $FF);
      g:=(p0[x] shr 8) and $FF;
      r:=(p0[x] shr 16) and $FF;

    dx:=round(x);
    dy:=round(y);
    rt:=sqrt(x*x+y*y);
    theta := ArcTan2(dx,dy)-rt;
    x1 :=round(temp.Width/2+( rt * Cos(theta)));
    y1 :=round(temp.height/2+( rt * Sin(theta)));


    if x1>temp.width-1 then x1:=x1-temp.Width;
    if y1>temp.height-1 then y1:=y1-temp.height;

    x1:=ensurerangeiiii(x1,0,temp.width-1);
    y1:=ensurerangeiiii(y1,0,temp.height-1);
    SS:=bmp.PixelPtr[x1,y1];
    SS^ := $FF000000 + r shl 16 + g shl 8 + b;
//    inc(SS);
    end;
  end;
  temp.free;
  bmp.Changed;
end;


 // == HLS / RGB =======================================================
  //
  // Based on C Code in "Computer Graphics -- Principles and Practice,"
  // Foley et al, 1996, p. 596.

function HSLtoRGB(H, S, L: Integer): TColor32;
var
  V, M, M1, M2, VSF: Integer;
begin
 Result := Color32(0, 0, 0, 0);
  if L <= $7F then
    V := L * (256 + S) shr 8
  else
    V := L + S - L * S div 255;
  if V <= 0 then
    Result := Color32(0, 0, 0, 0)
  else
  begin
    M := L * 2 - V;
    H := H * 6;
    VSF := (V - M) * (H shr 8) and $FF;
    M1 := M + VSF;
    M2 := V - VSF;
    case H shr 8 of
      0: Result := Color32(V, M1, M, 0);
      1: Result := Color32(M2, V, M, 0);
      2: Result := Color32(M, V, M1, 0);
      3: Result := Color32(M, M2, V, 0);
      4: Result := Color32(M1, M, V, 0);
      5: Result := Color32(V, M, M2, 0);
    end;
  end;
end;

{function Max(const A, B, C: Integer): Integer;
asm
      CMP       EDX,EAX
      CMOVG     EAX,EDX
      CMP       ECX,EAX
      CMOVG     EAX,ECX
end;}

{function Min(const A, B, C: Integer): Integer;
asm
      CMP       EDX,EAX
      CMOVL     EAX,EDX
      CMP       ECX,EAX
      CMOVL     EAX,ECX
end;}

procedure RGBtoHSL(RGB: TColor32; out H, S, L: Byte);
var   T1,T2,T3:Integer;
  R, G, B, D, Cmax, Cmin, HL: Integer;
begin
  R := (RGB shr 16) and $ff;
  G := (RGB shr 8) and $ff;
  B := RGB and $ff;
    T1:=Max(R, G);T2:=Max(R, B);T3:=Max(T1,T2);
  Cmax :=T3;
    T1:=Min(R, G);T2:=Min(R, B);T3:=Min(T1,T2);
  Cmin := T3;//Min(R, G, B);
  L := (Cmax + Cmin) div 2;

  if Cmax = Cmin then
  begin
    H := 0;
    S := 0
  end
  else
  begin
    D := (Cmax - Cmin) * 255;
    if L <= $7F then
      S := D div (Cmax + Cmin)
    else
      S := D div (255 * 2 - Cmax - Cmin);

    D := D * 6;
    if R = Cmax then
      HL := (G - B) * 255 * 255 div D
    else if G = Cmax then
      HL := 255 * 2 div 6 + (B - R) * 255 * 255 div D
    else
      HL := 255 * 4 div 6 + (R - G) * 255 * 255 div D;

    if HL < 0 then HL := HL + 255 * 2;
    H := HL;
  end;
end;


procedure Effect_IncDecRGB(clip:tbitmap32;r0,g0,b0:integer);
var
p0:Pcolor32array;
x,y: Integer;
r,g,b: Word;
//h1,h2,h3:integer;
SS:Pcolor32;
begin
  SS := @clip.Bits[0];

  for y:=0 to clip.Height-1 do
  begin
    p0:=clip.scanline[y];
    for x:=0 to clip.Width-1 do
    begin
      b:=(p0[x] and $FF);
      g:=(p0[x] shr 8) and $FF;
      r:=(p0[x] shr 16) and $FF;
    r := EnsureRangeiiii(r+r0, 0,255);
    g := EnsureRangeiiii(g+g0, 0, 255);
    b := EnsureRangeiiii(b+b0, 0, 255);
    SS^ := $FF000000 + r shl 16 + g shl 8 + b;
    inc(SS);
    end;
  end;
  clip.Changed;
end;

procedure Effect_IncDecHSL(clip:tbitmap32;h0,s0,l0:integer);
var
p0:Pcolor32array;
x,y{,a,r,g,b}: Integer;
//rd,gd,bd:single;
hd,sd,vd:byte;
//gray:integer;
SS:Pcolor32;
ps:TColor32; //singlepixel
    function FMod(X, Y: Extended): Extended;
    asm
            FLD     Y
            FLD     X
            FPREM
            FWAIT
    end;
begin
  SS := @clip.Bits[0];
  for y:=0 to clip.Height-1 do
  begin
    p0:=clip.scanline[y];
    for x:=0 to clip.Width-1 do
    begin
    RGBToHSL(p0[x],hd,sd,vd);
    hd := EnsureRangeiiii(hd + (h0), -180, 180);
    sd := EnsureRangeiiii(sd + (s0), 0, 255);
    vd := EnsureRangeiiii(vd + (l0), 0, 255);
//    hd := FMod(hd + (h0/255), 1);

    ps:=HSLtoRGB(hd,sd,vd);
    SS^ := ps;
    inc(SS);
    end;
  end;
  clip.Changed;
end;

procedure Effect_Dither(src: TBitmap32);
Const
     MY = 500;
     RP = 0.2989;
     GP = 0.5866;
     BP = 1 - RP - GP;
var
  x, y: Integer;
  p0: PColor32array;
  R, G, B: Integer;
  L0, L1: array[-1..MY+1] of Integer;
  light: Integer;
SS:Pcolor32;
//h1,h2,h3:integer;
begin
  SS := @src.Bits[0];
  for y:=0 to src.Height-1 do
  begin
    p0:=src.scanline[y];
    L0 := L1;
    FillChar(L1, SizeOf(L1), #0);
    for x:=0 to src.Width-1 do
    begin
      b:=(p0[x] and $FF);
      g:=(p0[x] shr 8) and $FF;
      r:=(p0[x] shr 16) and $FF;
       light := Trunc(r * rp+ g * gp + b * bp);
       light := light + L0[Y];
      if light > 255 then
      begin
      SS^ := $FF000000 + 255 shl 16 + 255 shl 8 + 255;
          inc(SS);
      light:=light-255;
      end else
      begin
      SS^ := $FF000000 + 0 shl 16 + 0 shl 8 + 0;
          inc(SS);
      end;
          Inc(L0[x+1], light div 4);
          Inc(L1[x-1], light div 4);
          Inc(L1[x  ], light div 4);
          Inc(L1[x+1], light - 3*(light div 4) );
     end;
   end;
   src.changed;
end;





{
procedure Effect_FFT(Dst,Src:TBitmap32);
var
  x, y: Integer;
  SS: PColor32;
  ps: pcolor32array;
  C, R, G, B: Integer;
  D: double;
  x2, f2        : pflt_array;
  nbr_points  : longint;
  fft         : TFFTReal;
  h1,h2,h3:integer;
begin
  Dst.Assign(Src);

  nbr_points := dst.height*dst.width; //power of 2;
  GetMem(x2, nbr_points * sizeof_flt);
  GetMem(f2, nbr_points * sizeof_flt);
  fft := TFFTReal.Create(nbr_points);

  SS := @src.Bits[0];
    h1:=$0000FF; //0
    h2:=$00FF00; //1
    h3:=$FF0000; //2

  for y:=0 to src.height-1 do begin
   ps:=src.scanline[y];
   for x:=0 to src.width-1 do begin
      r:=(ps[x] and $FF);
      g:=(ps[x] shr 8) and $FF;
      b:=(ps[x] shr 16) and $FF;

       D:= trunc(r);
       x2^[x+y*src.height] :=D;
      end;
    end;

//    fft.do_fft(f2,x2);

  fft.do_fft(f2, x2);
//  fft.do_ifft(f2, x2);
  fft.rescale(x2);

    with dst do
    for y := 0 to Height do begin
      for x := 0 to Width do begin
       c := ensurerange(round(f2^[x+y*height]),0,255);

      SS^ := $FF000000 + c shl 16 + c shl 8 + c;
      inc(SS);
      end;
    end;

//  FreeMem(x2);
//  FreeMem(f2);
  fft.free;

  //  FreeMem(x2);
//  FreeMem(f2);

end;
}

//////
{procedure Effect_FFT(Src:TBitmap32);
const CFmt='(%7.2f,%7.2f)';

var
  NC: integer;
  Cinp,Cout,CScr,CInv: array of tComplex;
  x, y: Integer;
  DstPColor: PColor32;
  C, R, G, B: Integer;
  D: double;
  a,f: double;
  dst:tbitmap32;
  w,h:integer;
begin
  w:=src.width;
  h:=src.height;
  dst:=tbitmap32.create;
  dst.width:=w;
  dst.height:=h;

  Dst.Assign(Src);
  NC := dst.height*dst.width;

  SetLength(Cinp,NC);
  SetLength(Cout,NC);
  SetLength(CScr,NC);
  SetLength(CInv,NC);

  f:=1;

  with dst do
    for y := 0 to Height do begin
      for x := 0 to Width do begin
       DstPColor:= PixelPtr[x, y];
       C := DstPColor^;
       r := (C and $0000FF);
       g := (C and $00FF00) shr 8;
       b := (C and $FF0000) shr 16;
       D:= trunc(r);
      with Cinp[x+y*height] do
      begin a := 2*pi*f*d/nc; r := cos(a); i := sin(a) end;
     end;
   end;

  CFTG(Cinp,Cout,CScr,+NC);
//  CFTG(Cout,CInv,CScr,-NC);

{  Memo1.Lines.Add(Format('NC=%d',[NC]));
  s := 'Cinp: ';
  For j := 0 to NC-1 do with Cinp[j] do begin
    if j>0 then s := s+', ';
    s := s + Format(CFmt,[r,i]);
    end;
  Memo1.Lines.Add(s);
  s := 'Cout: ';
  For j := 0 to NC-1 do with Cout[j] do begin
    if j>0 then s := s+', ';
    s := s + Format(CFmt,[r,i]);
    end;
  Memo1.Lines.Add(s);
  s := 'Cinv: ';
  For j := 0 to NC-1 do with Cinv[j] do begin
    if j>0 then s := s+', ';
    s := s + Format(CFmt,[r,i]);
    end;
  Memo1.Lines.Add(s);
  Memo1.Lines.Add('');
}
{
  nbr_points := dst.height*dst.width; //power of 2;
  GetMem(x2, nbr_points * sizeof_flt);
  GetMem(f2, nbr_points * sizeof_flt);
  fft := TFFReal.Create(nbr_points);

    with dst do
    for y := 0 to Height do begin
      for x := 0 to Width do begin
       DstPColor:= PixelPtr[x, y];
       C := DstPColor^;
       r := (C and $0000FF);
       g := (C and $00FF00) shr 8;
       b := (C and $FF0000) shr 16;
       D:= trunc(r);
       x2^[x+y*height] :=D;
      end;
    end;

//    fft.do_fft(f2,x2);

  fft.do_fft(f2, x2);
//  fft.do_ifft(f2, x2);
  fft.rescale(x2);

    with dst do
    for y := 0 to Height do begin
      for x := 0 to Width do begin
       DstPColor:= PixelPtr[x, y];
       D := f2^[x+y*height];
       DstPColor^:=$FF000000 + round(D) shl 16 + 0 shl 8 + 0;;
      end;
    end;

//  FreeMem(x2);
//  FreeMem(f2);
  fft.free;

  //  FreeMem(x2);
//  FreeMem(f2);

end;
}




end.
