unit im_HistogramFRM;

{*********************************************}
{  This unit is a part of ImageE              }
{  Copyright  2003-2004 R.Geurts             }
{  See Readme.txt for licence information     }
{*********************************************}

interface

uses
  Windows, Messages, SysUtils, Classes,
  Graphics, Controls, Forms,
  Dialogs, StdCtrls, Buttons,
  GR32,GR32_Image,im_basic,math, ExtCtrls;

type
  THistogramCTRLForm = class(TForm)
    Image1: TImage;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    PROCEDURE UpdateDisplay;
    procedure CreateHistoStretchBitmap(dst:TBitmap32);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormDestroy(Sender: TObject);
  private
    OriginalRangeLeft:  INTEGER;
    ScaleFactor      :  DOUBLE;
    StretchFactor    :  DOUBLE;
    TailPercentLeft  :  DOUBLE;
    TailPercentRight :  DOUBLE;
{    ShowDesignHeight :  INTEGER;
    ShowDesignWidth  :  INTEGER;}

  public
    dst:tbitmap32;
    src:tbitmap32;
  end;

var
  HistogramCTRLForm: THistogramCTRLForm;
  stretchenabled:boolean;
implementation

uses im_histogramlib, im_MainFRM, im_D5;

{$R *.dfm}

procedure THistogramCTRLForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  histogramctrlformX := histogramctrlform.left;
  histogramctrlformY:= histogramctrlform.top;
  DoSaver;
  Application.ProcessMessages;
end;
procedure THistogramCTRLForm.FormDestroy(Sender: TObject);
begin
  dst.free;
  src.free;
end;
procedure THistogramCTRLForm.FormCreate(Sender: TObject);
begin
  dst:=tbitmap32.create;
  src:=tbitmap32.create;
  TailPercentLeft  :=  1.0;
  TailPercentRight := 99.0;
  StretchFactor    :=  1.0;
  histogramctrlform.left := histogramctrlformX;
  histogramctrlform.top := histogramctrlformY;
end;

procedure THistogramCTRLForm.FormShow(Sender: TObject);
begin
  src.Assign(dst);
  updatedisplay;
  histogramctrlform.left := histogramctrlformX;
  histogramctrlform.top := histogramctrlformY;
{histogramctrlform.top:=mainform.toolbar1.height+45;
histogramctrlform.left:=mainform.clientwidth-histogramctrlform.Width-15;}
end;

PROCEDURE THistogramctrlform.UpdateDisplay;
  VAR
    OriginalHistogram :  THistogram;
    OriginalRangeRight:  INTEGER;
    OriginalRange     :  INTEGER;
//    StretchedBitmap   :  TBitmap32;
//    StretchedHistogram:  THistogram;
    StretchedRange    :  INTEGER;

  PROCEDURE ShowHistogram(CONST Histogram:  THistogram;
                          CONST Image:  TImage);
    CONST
      clSkyBlue    = TColor($F0CAA6);   // RGB:  166 202 240
    VAR
      Kurtosis         :  DOUBLE;
      Maximum          :  BYTE;
      Mean             :  DOUBLE;
      Median           :  BYTE;
      Minimum          :  BYTE;
      Mode             :  BYTE;
      N                :  INTEGER;
      Skewness         :  DOUBLE;
      StandardDeviation:  DOUBLE;
  BEGIN
    Image.Canvas.Brush.Color := clwhite;
    Image.Canvas.FillRect(Rect(0, 0, Image.Width, Image.Height));

    Histogram.Draw(Image.Canvas);
    Histogram.GetStatistics(N, Minimum, Maximum,
       Mode, Median,
       Mean, StandardDeviation,
       Skewness, Kurtosis);


{    LabelStats.Caption := 'Statistics:  ' +
      Format('%d / %d / %d / %d / %.1f / %.1f / %.1f',
      [Minimum,Maximum,Mode,Median,Mean,StandardDeviation, Kurtosis-3.0])
}  END {UpdateHistogram};

BEGIN
  Screen.Cursor := crHourGlass;
  TRY
    OriginalHistogram := THistogram.Create;
    TRY
      // Ask for cpRed, but cpGreen or cpBlue will give identical results
      // since R = G = B for shades of gray
      GetHistogram(cpRed, dst, OriginalHistogram);
      ShowHistogram(OriginalHistogram, Image1);

      OriginalRangeLeft  := OriginalHistogram.GetPercentileLevel(TailPercentLeft);
      OriginalRangeRight := OriginalHistogram.GetPercentileLevel(TailPercentRight);
      OriginalRange      := OriginalRangeRight - OriginalRangeLeft;

      // Normally, StretchFactor is 1.00 and RangeStretched = 255 (full width)
      StretchedRange := OriginalRange + ROUND( StretchFactor*(255 - OriginalRange) );

      IF   OriginalRange = 0.0
      THEN ScaleFactor := 1.0
      ELSE ScaleFactor := StretchedRange / OriginalRange;

      if stretchenabled=true then
      begin
      createhistostretchbitmap(dst);
      GetHistogram(cpRed, dst, OriginalHistogram);
      ShowHistogram(OriginalHistogram, Image1);
      end;
    FINALLY
      OriginalHistogram.Free
    END
  FINALLY
    Screen.Cursor := crDefault
  END
END {UpdateDisplay};

procedure THistogramCTRLForm.CreateHistoStretchBitmap(dst:TBitmap32);
  VAR
    i,j               :  INTEGER;
{    Intensity         :  INTEGER;}
    OriginalRow       :  Pcolor32array;
{    StretchedRow      :  Pcolor32array;}
    h1,h2,h3:integer;
    SS:Pcolor32;
    ri,gi,bi:integer;
    rw,gw,bw:word;
begin
    h1:=$0000FF; //0
    h2:=$00FF00; //1
    h3:=$FF0000; //2
  // Create stretched bitmap
  SS := @dst.Bits[0];

  FOR j := 0 TO dst.Height-1 DO
  BEGIN
    OriginalRow  := dst.Scanline[j];
    FOR i := 0 TO dst.Width-1 DO
    BEGIN
      // Use the R component, but R = G = B for shades of gray
      bi:= originalrow[i] and h1;
      gi:= (originalrow[i] and h2) shr 8;
      ri:= (originalrow[i] and h3) shr 16;
      // Calculate new intensity level
      ri := ROUND(ScaleFactor*(rI - OriginalRangeLeft) );
      gi := ROUND(ScaleFactor*(gI - OriginalRangeLeft) );
      bi := ROUND(ScaleFactor*(bI - OriginalRangeLeft) );

      IF   rI < 0
      THEN rI := 0
      ELSE
        IF   ri > 255
        THEN rI := 255;
      IF   gI < 0
      THEN gI := 0
      ELSE
        IF   gi > 255
        THEN gI := 255;
      IF   bI < 0
      THEN bI := 0
      ELSE
        IF   bi > 255
        THEN bI := 255;

rw:=ri;
gw:=gi;
bw:=bi;
//rw,gw,bw
      SS^ := $FF000000 + rw shl 16 + gw shl 8 + bw;
      inc(ss);
    END
    END
END {CreateHistoStretchBitamp};










end.
